#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/11/13 13:50
# @Author  : streamer
# @File    : loader.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/11/13   CREATE
import numbers
from operator import itemgetter
from celery.utils.imports import symbol_by_name
from celery.app import backends

loader_by_name = symbol_by_name


def without_none_values(iterable):
    """Return a copy of ``iterable`` with all ``None`` entries removed.

    If ``iterable`` is a mapping, return a dictionary where all pairs that have
    value ``None`` have been removed.
    """
    try:
        return {k: v for k, v in iterable.items() if v is not None}
    except AttributeError:
        return type(iterable)((v for v in iterable if v is not None))


def build_component_list(comp_info, custom=None):
    """Compose a component list from a { class: order } dictionary."""

    if comp_info is None:
        comp_info = {}

    def _validate_values(comp_dict):
        """Fail if a value in the components dict is not a real number or None."""
        for name, value in comp_dict.items():
            if value is not None and not isinstance(value, numbers.Real):
                raise ValueError(
                    f"Invalid value {value} for component {name}, "
                    "please provide a real number or None instead"
                )

    if custom is not None:
        comp_info.update(custom)

    _validate_values(comp_info)
    comp_info = without_none_values(comp_info)
    return [k for k, v in sorted(comp_info.items(), key=itemgetter(1))]


def item_backend_by_url(url=None, app_loader=None):
    # Return item backend by url
    # TODO 应该写在一个专门的文件里边的
    from celery.app import backends
    item_backend_aliases = {
        'mongodb': 'spider.base.easy_item_backend: ItemMongoBackend',
    }
    app_loader = app_loader or backends.current_app.loader_by_name

    # 保存原始的配置
    origin_backend_settings = app_loader.override_backends.copy()
    origin_aliases = backends.BACKEND_ALIASES.copy()
    origin_unknown_item_backend = backends.UNKNOWN_BACKEND

    # 更换成item backend 的配置
    app_loader.override_backends = backends.current_app.conf.get('item_backend', {})
    backends.UNKNOWN_BACKEND = """
Unknown item backend: {0!r}.  Did you spell that correctly? ({1!r})
"""
    backends.BACKEND_ALIASES.update(item_backend_aliases)

    # 通过url 获取item backend
    item_backend, url = backends.by_url(url, loader_by_name)

    # 还原配置
    app_loader.override_backends = origin_backend_settings
    backends.BACKEND_ALIASES = origin_aliases
    backends.UNKNOWN_BACKEND = origin_unknown_item_backend

    return item_backend, url



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/11/9 15:48
# @Author  : streamer
# @File    : easy_item_backends.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/11/9   CREATE
import threading
import time
from datetime import datetime
import pytz
from bson import ObjectId

from celery.exceptions import BackendStoreError
from celery.utils.serialization import raise_with_context
from celery.utils.time import get_exponential_backoff_interval
from spider.utils.trackref import get_create_time
from celery.backends.mongodb import MongoBackend


class TrimBackendMeta(type):
    # TODO 从celery.backends.mongodb import MongoBackend 修剪成base backend，然后删除不需要的代码

    def __call__(cls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)

        # 删除从mongodb 继承的多余属性
        trim_attrs = [
            'taskmeta_collection',
            'groupmeta_collection',
            '_store_result',
            '_get_task_meta_for',
            '_save_group',

            '_restore_group',
            '_delete_group',
            '_forget',
            'cleanup',
            'collection',
            'group_collection'
        ]
        for attr in trim_attrs:
            if hasattr(instance, attr):
                try:
                    delattr(instance, attr)
                except AttributeError:
                    at = getattr(instance, attr)
                    del at

        return instance


class ItemMongoBackend(MongoBackend, metaclass=TrimBackendMeta):

    def __init__(self, app, **kwargs):
        super().__init__(app, **kwargs)
        # TODO 读取 settings 中添加 item -> collection 映射的配置
        # TODO 最好添加一个锁，防止多线程同时修改，
        self.collections = {}

    def _get_collection(self, col_name: str):
        # if col_name not in self.collections:
        #     self.collections[col_name] = self.database[col_name]
        # return self.collections[col_name]
        return self.database[col_name]

    def _get_item_meta(self, item):
        # TODO make mongodb ObjectId from obj to str
        obj_id = ObjectId()
        item['_id'] = str(obj_id)
        item_meta = dict(item)
        # TODO 最好在基类里边自动设置item 的create_time, update_time等 追踪时间
        # TODO 可以从settings 中读取 时区
        time_zone = pytz.timezone('Asia/Shanghai')

        item_meta['update_time'] = datetime.now(time_zone)
        create_time_stamp = get_create_time(item)
        utc_time = datetime.utcfromtimestamp(create_time_stamp)
        item_meta['create_time'] = utc_time.replace(tzinfo=pytz.utc).astimezone(time_zone)
        return item_meta

    def get_col(self, name):
        return self._get_collection(name)

    def _store_item(self, item):
        col_name = item.__class__.__name__
        item_meta = self._get_item_meta(item)
        col = self._get_collection(col_name)
        col.insert_one(item_meta)
        return item_meta

    def store_item(self, item):

        retries = 0
        while True:
            try:
                item_meta = self._store_item(item)
                return item_meta
            except Exception as exc:
                if self.always_retry and self.exception_safe_to_retry(exc):
                    if retries < self.max_retries:
                        retries += 1

                        # get_exponential_backoff_interval computes integers
                        # and time.sleep accept floats for sub second sleep
                        sleep_amount = get_exponential_backoff_interval(
                            self.base_sleep_between_retries_ms, retries,
                            self.max_sleep_between_retries_ms, True) / 1000
                        self._sleep(sleep_amount)
                    else:
                        raise_with_context(
                            BackendStoreError("failed to store result on the backend", ),
                        )
                else:
                    raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/11/20 15:10
# @Author  : streamer
# @File    : __init__.py.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/11/20   CREATE
from .easy_item import BaseItem
from .easy_obj import EasyObj


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/10/25 10:47
# @Author  : streamer
# @File    : response.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/10/25   CREATE
import requests

Response = requests.Response

# class Response(EasyObj):
#     pass



# -*- coding:utf-8 -*-
# Time:   2023/10/30 1:41
# Author: streamer
# Email:  streamer5538@gmail.com
from typing import Any

from .base import MiddlewareManger
from spider.base.easy_chain import SkipOnErrChain


class PipelineMiddlewareManager(MiddlewareManger):
    component_name = 'pipeline_middlewares'

    def _add_middleware(self, pipe: Any) -> None:
        if hasattr(pipe, 'process_item') and callable(pipe.process_item):
            self.methods['process_item'].append(pipe.process_item)

    def process_item(self, item: Any) -> Any:
        # process crawl item
        chain = SkipOnErrChain()
        for method in self.methods['process_item']:
            chain.add_callback(method)

        chain.callback(item)
        return chain.result





# -*- coding:utf-8 -*-
# Time:   2023/10/30 1:44
# Author: streamer
# Email:  streamer5538@gmail.com

from .base import ComponentManger
from typing import Any, Dict, Union, Callable
from celery import Task
import threading
from spider.utils import get_full_name, get_obj_func
import logging

logger = logging.getLogger(__name__)
"""
将spider_parser 分解成两个类:
    A、parser_class 保存解析方法和加载解析方法
    B、parser_bridge 维护parser 和spider 的关系，包括保存task-callback, register_crawl，以及在register task 时保存到callback_str
    C、目的： 
        --根据callback 取出task 进行运行， 如果没有task，取出base_task作为task 运行  --> 需要一个字典以callback 为键，task 为值保存关系，命名为parse_task_map
        --需要根据callback_str 取出对应的callback，让base_task 可以实例化进行        --> 需要一个字典以callback_str 为键，callback 为值保存关系, 命名为parsers
        --现在有两种情况; 
            第一种是celery 初始化前加载的parser_callback 都要生成唯一对应的task，方便直接通过task.delay 的调用；
            第二种是动态加载的parser_callback，由于这时候celery 无法注册task了，所以需要在parsers字典中保存信息，方便base_task 加载parse_callback;
                --如何应对这种情况呢：
                    1、最顶级的方法我认为应该修改celery 的task 注册机制，让celery 在运行中也可以注册task；好处是不需要多维护一个parsers 的字典; 
                坏处是需要修改到celery 中的源码，且可能引发线程问题或者不知名问题
                    2、使用双字典进行数据的保存---- celery 中有registry 用来保存task, 只需要将函数的名称替换成full_name就行了（从解析器里边加载的parse 不会定义名字，
                不用担心重载）；只需要维护一个callback_str 的字典就可以了,可以通过callback 来获取名称
        
"""


class SpiderParserManager(ComponentManger):
    """

    parser middleware manager and maintenance relationship between parser and task
    """
    component_name = 'spider_parser'

    def __init__(self, components, spider) -> None:
        self.parsers: Dict[str, Callable] = dict()
        super().__init__(components, spider=spider)
        self.lock = threading.Lock()
        self.task_maker = spider.task
        for cp_name, cp_obj in components.items():
            # print(f'init: component:{cp_obj}, name:{cp_name}')
            self.add_parse_func(cp_obj)

    def set(self, parser_func: Callable, name: str) -> None:
        self.parsers[name] = parser_func

    def get(self, name: str, default) -> Union[Callable, None]:
        return self.parsers.get(name, default)

    def delete(self, name: str) -> None:
        del self.parsers[name]

    def get_thread_safe(self, name: str, default=None) -> Union[Callable, None]:
        with self.lock:
            return self.get(name, default)

    def contains_parser(self, parser: Union[str, Callable]) -> bool:
        """check if parser in parsers, other set parser in parsers"""
        with self.lock:
            if parser in self.parsers.keys():
                return True
            else:
                self.set(parser, parser)
                return False

    def add_parse_func(self, parser_obj: Any) -> None:
        # FIXME 这边需要对parse 做严格的数据限定,不然多线程运行时, spider 类的数据出错
        # TODO 应该设置可以从配置中读取对应的parser, 由于项目需要, 暂时设定从spider类中读取
        parse_funcs = get_obj_func(parser_obj, start_with='parse')
        for parse_func in parse_funcs:
            parse_func_name = get_full_name(parse_func)
            self.set(parse_func, parse_func_name)
            self.task_maker(parse_func, name=parse_func_name)

    def update_settings(self, settings):
        """可能出现在初始化之后 parser_manager 后在加载配置的情况，所以要添加一个更新的方法"""
        new_components = self.load_components(settings)
        for component in new_components:
            if component is not None and component not in self.components:
                self.components[component] = component
                self.add_parse_func(component)

    @classmethod
    def _get_components_from_settings(cls, settings):
        return settings.get(cls.component_name) or settings.get(cls.component_name.upper()) or []

    @classmethod
    def from_settings(cls, settings, spider=None, **kwargs):
        components = cls.load_components(settings)
        return cls(components, spider=spider)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/10/30 10:05
# @Author  : streamer
# @File    : easy_chain.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/10/30   CREATE
from typing import (
    Callable,
    Any,
    cast,
    TypeVar,
    Generic,
    List,
    Tuple
)
import logging

logger = logging.getLogger(__name__)

T = TypeVar('T', bound='BaseChain')


def _execute_callback(func: Callable, *args: Any, **kwargs: Any):
    try:
        result = func(*args, **kwargs)
    # TODO set my exception type
    except Exception as e:
        logger.info('when execute callback {} counter exception: {}'.format(func.__name__, e))
        result = e

    return result


class BaseChain(Generic[T]):
    """spider chain class for callbacks and err_callbacks"""

    def __init__(self) -> None:
        self.callbacks: List[Tuple[Callable, Any, Any]] = list()
        self.err_callbacks: List[Tuple[Callable, Any, Any]] = list()
        self._fired: bool = False
        self._result: Any = None

    def add_callback(self, callback: Callable, *args: Any, **kwargs: Any) -> T:
        if self._fired:
            self._result = callback(self._result, *args, **kwargs)
        else:
            self.callbacks.append((callback, args, kwargs))
        return self

    def add_err_callback(self, err_callback: Callable, *args: Any, **kwargs: Any) -> T:
        if self._fired and isinstance(self._result, Exception):
            self._result = err_callback(self._result, *args, **kwargs)
        else:
            self.err_callbacks.append((err_callback, args, kwargs))
        return self

    def callback(self, *args: Any, **kwargs) -> Any:
        # FIXME 为了允许在启动链式反应时可以进行动态传参,需要先进行一次回调,这样可以防止开始运行时的动态参数和后续函数的动态函数不会相互污染
        # FIXME 在找到更好的处理方式前，暂定处理方式为在callback 中进行第一次的回调
        # FIXME callback func must have value args? need fix

        if self.callbacks:
            func, func_args, func_kwargs = self.callbacks.pop(0)
            args = args + func_args
            kwargs.update(func_kwargs)
            result = _execute_callback(func, *args, **kwargs)
            result = self._fire_callbacks(result)
            self._result = result
            self._fired = True

    def _fire_callbacks(self, result: Any) -> None:
        raise NotImplementedError

    @property
    def result(self) -> Any:
        return self._result


class PairErrBackChain(BaseChain):
    """chain class that each callback has a corresponding err_callback"""

    def _fire_callbacks(self, result: Any) -> None:

        # FIXME HACK 关于错误的处理逻辑：通用处理方式，报错时直接从err_callbacks 出队直接回调
        # FIXME HACK 由于没有对callback 和err back 做强制对应, 可能会导致调用时出现差错
        while self.callbacks or self.err_callbacks:
            if isinstance(result, Exception) and self.err_callbacks:
                err_callback, args, kwargs = self.err_callbacks.pop(0)
                result = err_callback(result, *args, **kwargs)
            elif self.callbacks:
                callback, args, kwargs = self.callbacks.pop(0)
                result = _execute_callback(callback, result, *args, **kwargs)
            else:
                break
        return result


class UnifiedErrBackChain(BaseChain):
    """chain class where all callbacks correspond to the same err_callback"""

    def _fire_callbacks(self, result: Any) -> None:
        # FIXME HACK 没有对err_back 有且只有一个限定, 也没有更换err_back 的方法
        while self.callbacks:
            if isinstance(result, Exception) and self.err_callbacks:
                err_callback, args, kwargs = self.err_callbacks[0]
                result = err_callback(result, *args, **kwargs)
            elif self.callbacks:
                callback, args, kwargs = self.callbacks.pop(0)
                result = _execute_callback(callback, result, *args, **kwargs)
            else:
                break


class SkipOnErrChain(BaseChain):
    """chain class that ignores all exceptions generated from callbacks"""

    def _fire_callbacks(self, result: Any) -> None:
        while self.callbacks:
            callback, args, kwargs = self.callbacks.pop(0)
            try:
                new_args = (result,) + args
                result = callback(*new_args, **kwargs)
            except Exception as e:
                logger.warning(f'when {self.__class__.__name__} executes callback {callback.__name__} '
                               f'counter exception: {e}')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/10/25 10:40
# @Author  : streamer
# @File    : __init__.py.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/10/25   CREATE
from requests import request
from selenium import webdriver
from .request import Request
from .response import Response


def download_page(request: Request):
    pass


def request_page(url, method='GET', headers=None, data=None, *args, **kwargs):
    if headers is None:
        headers = {}
    if data is None:
        data = {}
    return request(method, url, headers=headers, data=data, *args, **kwargs)


def request_page_by_selenium(url):
    # TODO 不要一直打开关闭chromedriver，应该做成chromedriver pool
    # step 1:set chromedriver options
    options = webdriver.ChromeOptions()
    options.add_argument('--lang=en')
    options.add_argument('--headless')
    options.add_argument('--disable-gpu')
    options.add_argument('--no-sandbox')
    options.add_argument('--disable-dev-shm-usage')
    options.add_argument('log-level=3')
    options.add_argument('--disable-blink-features=AutomationControlled')  # 谷歌浏览器去掉访问痕迹
    options.add_argument(
        "--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
        "Chrome/100.0.4896.127 Safari/537.36")
    options.add_argument("--window-size=1920,1050")  # 专门应对无头浏览器中不能最大化屏幕的方案
    # step 2: init chromedriver
    driver = webdriver.Chrome(
        # executable_path=chromedriver_path,
        options=options
    )
    # step 3: access the specified url address
    driver.get(url)
    # step 4: get the page
    resp = Response()
    resp._content = driver.page_source
    resp.url = driver.current_url
    resp.status_code = 200
    driver.quit()
    return resp





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/11/3 16:14
# @Author  : streamer
# @File    : __init__.py.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/11/3   CREATE
from typing import Any, List, Union, Callable


def get_full_name(obj: Any) -> str:
    return f'{obj.__module__}.{obj.__qualname__}'


def get_obj_func(
        obj: Any, func_name: str = None, start_with: str = None, end_with: str = None, contains: str = None
) -> List[Union[Callable, None]]:
    if func_name is not None and hasattr(obj, func_name) and callable(getattr(obj, func_name)):
        return [getattr(obj, func_name)]
    if start_with is not None:
        return [getattr(obj, func_name) for func_name in dir(obj) if
                func_name.startswith(start_with) and callable(getattr(obj, func_name))]
    if end_with is not None:
        return [getattr(obj, func_name) for func_name in dir(obj) if
                func_name.endswith(end_with) and callable(getattr(obj, func_name))]
    if contains is not None:
        return [getattr(obj, func_name) for func_name in dir(obj) if
                contains in func_name and callable(getattr(obj, func_name))]
    return []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/10/25 10:46
# @Author  : streamer
# @File    : request.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/10/25   CREATE
from spider.base import EasyObj
from spider.utils import get_full_name


class Request(EasyObj):
    def __init__(self, url=None, callback=None, method='GET', headers=None, data=None, meta=None, **kwargs):
        self.url = url
        self.method = method
        self.headers = headers
        self.data = data
        self.meta = meta
        self.callback = callback
        EasyObj.__init__(self, **kwargs)

    def import_callback(self, module):
        # path = f'{module}.{self.callback}'
        # # module = symbol_by_name(module)
        # # self.callback = getattr(module, self.callback)
        # self.callback = load_function(path)
        # return self
        pass

    @property
    def str_callback(self):
        callback = ''
        if 'callback' in self.__dict__.keys() and callable(self.callback):
            # callback = self.__dict__.pop('callback')
            callback = get_full_name(self.callback)
        return callback

    @property
    def serialize(self):
        data = {
            k: v for k, v in self.__dict__.items()
            if v and (not (k.startswith('_') or callable(getattr(self, k))))
        }
        # if 'callback' in self.__dict__.keys():
        #     del data['callback']
        # data.update({
        #     k: v for k, v in self.__dict__.items()
        #     if v and (not (k.startswith('_') or callable(getattr(self, k))))
        # })
        return data

    @property
    def serialize_with_callback_str(self):
        data = self.serialize
        data['callback'] = self.str_callback
        return data





# -*- coding:utf-8 -*-
# Time:   2023/10/30 1:41
# Author: streamer
# Email:  streamer5538@gmail.com
from .base import MiddlewareManger
from spider.network import Request, Response, download_page
from typing import Any, Callable, Union, cast
from spider.base.easy_chain import PairErrBackChain


class DownloaderMiddlewareManager(MiddlewareManger):
    component_name = 'downloader_middleware'

    def _add_middlewares(self, mw: Any) -> None:
        if hasattr(mw, 'process_request') and callable(mw.process_request):
            self.methods['process_request'].append(mw.process_request)
        if hasattr(mw, 'process_response') and callable(mw.process_response):
            self.methods['process_response'].appendleft(mw.process_response)
        if hasattr(mw, 'process_exception') and callable(mw.process_exception):
            self.methods['process_exception'].appendleft(mw.process_exception)

    def download(self, request: Request, download_func: Callable = None) -> Union[Request, Response, Exception]:
        if download_func is None:
            download_func = download_page

        def process_request(req: Request) -> Union[Request, Response]:

            for method in self.methods['process_request']:
                method = cast(Callable, method)
                resp = method(req)
                if resp is not None and not isinstance(resp, (Response, Request)):
                    # TODO set my value exception
                    raise TypeError(
                        f'Middleware {method.__qualname__} must return Response or Request, '
                        f'got {resp.__class__.__name__}'
                    )
                elif isinstance(resp, Request):
                    req = resp
                else:
                    return resp
            return download_func(request=req)

        def process_response(resp: Response) -> Union[Request, Response]:
            if resp is None:
                raise TypeError(f'Received None in Middleware[{self.component_name}] process_response')
            elif isinstance(resp, Request):
                return resp

            for method in self.methods['process_response']:
                method = cast(Callable, method)
                # FIXME this request never through process_request
                resp = method(request=request, response=resp)
                if not isinstance(resp, (Response, Request)):
                    # TODO set my exception type
                    raise TypeError(
                        f'Middleware {method.__qualname__} must return Response or Request,'
                        f'get {type(resp)}'
                    )
                if isinstance(resp, Request):
                    return resp
            return resp

        def process_exception(ex: Exception) -> Union[Request, Response, Exception]:
            for method in self.methods['process_exception']:
                method = cast(Callable, method)
                resp = method(request=request, exception=ex)
                if resp is not None and not isinstance(resp, (Response, Request)):
                    raise TypeError(
                        f'Middleware {method.__qualname__} must return None, Response or Request,'
                        f'got {type(resp)}'
                    )
                if resp:
                    return resp
            return ex

        download_chain = PairErrBackChain()
        download_chain.add_callback(process_request)
        download_chain.add_callback(process_response)
        download_chain.add_err_callback(process_exception)
        download_chain.callback(request)
        return download_chain.result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/6/12 05:23
# @Author  : streamer
# @File    : easy_item.py
# @Project : e_commerce
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/6/12   CREATE

from abc import ABCMeta
from collections.abc import MutableMapping
from copy import deepcopy
from pprint import pformat
from typing import Dict
from spider.utils.trackref import ObjectRef


class Attr(dict):
    """attribute"""


is_debug = False


def print_info(info):
    if is_debug:
        import pprint
        pprint.pprint(info)


class BaseModelMeta(ABCMeta):

    def __new__(mcs, class_name, bases, attrs):
        classcell = attrs.pop("__classcell__", None)
        new_bases = tuple(base._class for base in bases if hasattr(base, "_class"))
        print_info(f'super() type:{type(super())}, value:{super()}')
        _class = super().__new__(mcs, "x_" + class_name, new_bases, attrs)
        print_info(f'_class type:{type(_class)}, value:{_class}')
        fields = getattr(_class, "fields", {})
        print_info(f'fields type:{type(fields)}, value:{fields}')
        new_attrs = {}
        for n in dir(_class):
            v = getattr(_class, n)
            if isinstance(v, Attr):
                fields[n] = v
            elif n in attrs:
                new_attrs[n] = attrs[n]

        new_attrs["fields"] = fields
        new_attrs["_class"] = _class
        if classcell is not None:
            new_attrs["__classcell__"] = classcell
        return super().__new__(mcs, class_name, bases, new_attrs)


class BaseItem(MutableMapping, ObjectRef, metaclass=BaseModelMeta):
    fields: Dict[str, Attr]

    def __init__(self, *args, **kwargs):
        self._values = {}
        if args or kwargs:  # avoid creating dict for most common case
            for k, v in dict(*args, **kwargs).items():
                self[k] = v

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        self._values[key] = value

    def __delitem__(self, key):
        del self._values[key]

    def __getattr__(self, name):
        if name in self.fields:
            return self.__getitem__(name)
        else:
            return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if name in self.fields:
            self.__setitem__(name, value)
        else:
            super().__setattr__(name, value)

    def __len__(self):
        return len(self._values)

    def __iter__(self):
        return iter(self._values)

    __hash__ = ObjectRef.__hash__

    def keys(self):
        return self._values.keys()

    def __repr__(self):
        return pformat(dict(self))

    def copy(self):
        return self.__class__(self)

    def deepcopy(self):
        """Return a :func:`~copy.deepcopy` of this item."""
        return deepcopy(self)

    @property
    def id(self):
        if hasattr(self, 'id') or 'id' in self.__dict__.keys():
            return self.__dict__['id']
        elif hasattr(self, '_id') or '_id' in self.__dict__.keys():
            return self.__dict__['_id']
        else:
            # TODO  set my exception
            raise Exception(f'no id field in {self.__class__.__name__}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/11/13 13:54
# @Author  : streamer
# @File    : exceptions.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/11/13   CREATE


class ScrapyDeprecationWarning(Warning):
    """Warning category for deprecated features, since the default
    DeprecationWarning is silenced on Python 2.7+
    """

    pass





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/10/25 10:40
# @Author  : streamer
# @File    : __init__.py.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/10/25   CREATE

from .components import (
    MiddlewareManger,
    DownloaderMiddlewareManager,
    SpiderParserManager,
    PipelineMiddlewareManager
)
from spider.base.easy_item_backends import ItemMongoBackend
from functools import wraps, update_wrapper, cached_property
from collections.abc import Iterator
from celery import Celery, Task

from .base import BaseItem
from spider.network import Request
from spider.utils.loader import loader_by_name, item_backend_by_url
import logging

from .utils import get_full_name

logger = logging.getLogger(__name__)


class CeleryMixin(Celery):

    def __init__(self, main=None, loader=None, backend=None, amqp=None, events=None, log=None, control=None,
                 set_as_current=True, tasks=None, broker=None, include=None, changes=None, config_source=None,
                 fixups=None, task_cls=None, autofinalize=True, namespace=None, strict_typing=True, **kwargs):
        super().__init__(main, loader, backend, amqp, events, log, control, set_as_current, tasks, broker, include,
                         changes, config_source, fixups, task_cls, autofinalize, namespace, strict_typing, **kwargs)


class SpiderMixin:
    downloader_cls = 'spider.components.downloader.DownloaderMiddlewareManager'
    parser_cls = 'spider.components.parser.SpiderParserManager'
    pipeline_cls = 'spider.components.pipeline.PipelineMiddlewareManager'
    item_backend_cls = None

    # downloader: DownloaderMiddlewareManager = None
    # parser: SpiderParserManager = None
    # pipeline: PipelineMiddlewareManager = None

    def __init__(self, name=None, *, downloader_cls=None, parser_cls=None, pipeline_cls=None,
                 item_backend_cls=None, **kwargs):
        self.spider_name = name or self.__class__.__name__
        # super().__init__(name, **kwargs)
        self.downloader_cls = downloader_cls or self.downloader_cls
        self.parser_cls = parser_cls or self.parser_cls
        self.pipeline_cls = pipeline_cls or self.pipeline_cls
        self.item_backend_cls = item_backend_cls or self.item_backend_cls

        self._item_backend_cached = None

    @cached_property
    def downloader(self) -> DownloaderMiddlewareManager:
        return self._init_downloader()

    @cached_property
    def parser(self) -> SpiderParserManager:
        return self._init_parser()

    @cached_property
    def pipeline(self) -> PipelineMiddlewareManager:
        return self._init_pipeline()

    def _init_downloader(self):
        downloader_cls = loader_by_name(self.downloader_cls)
        return downloader_cls.from_settings(self.conf)

    def _init_parser(self):
        parser_cls = loader_by_name(self.parser_cls)
        parser = parser_cls.from_settings(self.conf, self)
        # self.load_task_from_parser(parser)
        return parser

    def _init_pipeline(self):
        pipeline_cls = loader_by_name(self.pipeline_cls)
        return pipeline_cls.from_settings(self.conf)

    def _crawl_flow(self, parse_func):
        """spider crawl flow"""

        @wraps(parse_func)
        # TODO HACK should make crawl args and kwargs to Request class
        def crawl(url, *args, **kwargs):

            result = {}
            items = []
            request = Request(url, *args, **kwargs)
            # TODO Add Features: Custom download methods can be loaded from settings
            resp = self.downloader.download(request=request, download_func=None)

            # TODO Add Features: Choose whether to parse web pages in this node from settings
            parse_result = parse_func(resp)

            # sure parser_result is generator
            if not (hasattr(parse_result, '__next__') and callable(getattr(parse_result, '__next__'))):
                if isinstance(parse_result, Iterator):
                    parse_result = (x for x in parse_result)
                else:
                    parse_result = (x for x in [parse_result])
            is_continued = True
            while is_continued:
                try:
                    item = next(parse_result)
                except StopIteration as e:
                    item = e.value  # make sure can get item when use return in parse_func
                    is_continued = False

                if isinstance(item, Request):
                    self.publish_crawl_task(item)
                # elif isinstance(item, Task):  # 貌似不会yield Task 的情况出现

                elif isinstance(item, BaseItem):
                    item = self.pipeline.process_item(item)
                    # TODO 添加可以根据settings或者opts 自动设定存储item的功能
                    item = self.store_item(item)
                    items.append(item)
                else:
                    # TODO setting my exception
                    logger.warning(f'methods[{parse_func.__qualname__}] should yield or return Request or Item'
                                   f', got %s' % type(item))

            # TODO 丰富日志的打印
            logger.info(f'crawl {len(items)} items in page[url="{url}"]')

            result['item_ids'] = [item.id for item in items]

            return result

        return crawl

    def crawl_task(self, *args, **kwargs):
        def wrapper(func):

            _crawl_flow = self._crawl_flow(func)
            new_args = (_crawl_flow,) + args[1:]
            update_wrapper(_crawl_flow, func)

            task = self.task(*new_args, **kwargs)

            return task

        if len(args) == 1 and callable(args[0]):
            return wrapper(args[0])
        else:
            return wrapper

    def _init_base_crawl_task(self):
        def _base_task(*args, **kwargs):
            request = Request(*args, **kwargs)
            parse_func_name = request.callback
            if parse_func := self.parser.get_thread_safe(parse_func_name):
                return self._crawl_flow(parse_func)(*args, **kwargs)
            else:
                # TODO  set my exception
                raise Exception(f'parse_func[{parse_func_name}] not found')

        task = self.task(_base_task, name='base_crawl')
        return task

    def publish_crawl_task(self, request: Request) -> None:
        # TODO 根据settings 在这里添加去重逻辑
        crawl_task = self.tasks.get(request.str_callback)
        if crawl_task is not None:
            crawl_task.delay(**request.serialize)
        else:
            self.parser.contains_parser(request.callback)
            crawl_task = self.base_crawl_task
            crawl_task.delay(**request.serialize_with_callback_str)

    def task(self, *args, **opts) -> Task:
        """celery class task"""

    @cached_property
    def base_crawl_task(self):
        if task := self.tasks.get('base_crawl'):
            return task
        else:
            # TODO Set my exception
            raise Exception('base_crawl task not found')

    def store_item(self, item):
        raise NotImplementedError

    @property
    def conf(self):
        """celery class conf"""
        raise NotImplementedError

    @property
    def tasks(self):
        """celery class tasks"""
        raise NotImplementedError


class CelerySpider(CeleryMixin, SpiderMixin):
    item_backend_cls = None

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)

    @property
    def _item_backend(self):
        if self._item_backend_cached is not None:
            return self._item_backend_cached
        return getattr(self._local, 'item_backend', None)

    @_item_backend.setter
    def _item_backend(self, item_backend):
        if item_backend.thread_safe:
            self._item_backend_cached = item_backend
        else:
            self._local.backend = item_backend

    @property
    def item_backend(self):
        if self._item_backend is None:
            self._item_backend = self._get_item_backend()
        return self._item_backend

    def _get_item_backend(self):
        # TODO 设置一个参数, 确定是否result 和item 是否使用同一个数据库,这样可以减少数据库的连接参数设置
        if self.conf.get('item_use_result_backend', True):
            self.conf.item_backend = self.conf.result_backend
        backend, url = item_backend_by_url(
            self.item_backend_cls or self.conf.item_backend,
            self.loader
        )
        return backend(app=self, url=url)

    def store_item(self, item):
        return self.item_backend.store_item(item)

    # def on_init(self):
    #     self._init_base_crawl_task()

    # def start(self, argv=None):
    #     self.parser.update_settings(self.conf)
    #     self._init_base_crawl_task()
    #     super().start(argv=argv)

    def worker_main(self, argv=None):
        self.parser.update_settings(self.conf)
        self._init_base_crawl_task()
        super().worker_main(argv=argv)



# -*- coding:utf-8 -*-
# Time:   2023/10/30 1:08
# Author: streamer
# Email:  streamer5538@gmail.com
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Union,
    Tuple, List
)
from collections import defaultdict, deque, OrderedDict
from spider.utils.loader import loader_by_name, build_component_list
import logging
import pprint

logger = logging.getLogger(__name__)


class ComponentManger(object):
    component_name = "component_manager"

    def __init__(self, components, spider=None) -> None:
        self.components: Dict[str, Any] = components
        self.spider = spider

        self.on_init()

    def on_init(self):
        """ Called when spider is init. """


    @classmethod
    def _get_components_from_settings(cls, settings):
        # print(f'{cls.component_name}: {settings.get(cls.component_name) or settings.get(cls.component_name.upper())}')
        return build_component_list(
            settings.get(cls.component_name) or settings.get(cls.component_name.upper())
        )

    @classmethod
    def load_components(cls, settings) -> Dict[str, Any]:
        cp_path = cls._get_components_from_settings(settings)
        components = OrderedDict()
        enabled = []
        for cls_path in cp_path:
            try:
                cp_cls = loader_by_name(cls_path)
                components[cls_path] = cp_cls
                enabled.append(cls_path)
            # TODO set my config error exception
            except Exception as e:
                logger.warning(e)
        logger.info(
            "Enabled %(componentname)ss:\n%(enabledlist)s",
            {
                "componentname": cls.component_name,
                "enabledlist": pprint.pformat(enabled),
            },
        )
        return components

    @classmethod
    def from_settings(cls, settings, spider=None, **kwargs):
        components = cls.load_components(settings)

        return cls(components, spider=spider)

    def _add_middleware(self, mw: Any) -> None:
        raise NotImplementedError


class MiddlewareManger(ComponentManger):
    component_name = "middleware_manager"

    def __init__(self, middlewares, spider) -> None:
        super().__init__(middlewares, spider)
        self.methods: Dict[
            str, Deque[Union[None, Callable, Tuple[Callable, Callable]]]
        ] = defaultdict(deque)

    def on_init(self):
        for mw in self.components.values():
            self._add_middleware(mw)

    @classmethod
    def from_settings(cls, settings, spider=None, **kwargs):
        return cls(*cls.load_components(settings))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2023/10/25 10:56
# @Author  : streamer
# @File    : easy_obj.py
# @Project : celery_spider
# @Software: PyCharm
# @History : 
# VERSION     USER      DATE         DESC
# v1.0.0      Streamer   2023/10/25   CREATE
# TODO 现在暂时没时间完善下载器和下载中间件，所以简单写个obj和dict 可以相互转换的类
class EasyObj:
    """
    简单的对象类型, 用于存储简单的对象信息
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    def to_dict(self):
        return vars(self).copy()

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__})'

    def __str__(self):
        return f'{self.__class__.__name__}({self.__dict__})'


